from dna_features_viewer import BiopythonTranslator

class DiffRecordTranslator(BiopythonTranslator):

    ignored_features_types = ("diff_equal",)

    @staticmethod
    def compute_feature_color(f):
        return dict(
            diff_delete="#E76F51",  # RED
            diff_insert="#2A9D8F",  # GREEN
            diff_replace="#E9C46A",  # YELLOW
            diff_change="#F4A261",  # ORANGE
            diff_reverse="white",
            diff_transpose="white",
        ).get(f.type, "white")

    @staticmethod
    def compute_feature_box_linewidth(f):
        return 1 if f.type.startswith("diff_") else 0

    @staticmethod
    def compute_feature_fontdict(f):
        return {"fontsize": 12 if f.type.startswith("diff_") else 7}


from copy import deepcopy

import networkx as nx
import Levenshtein

from ..Location import Location

from .DiffBlock import DiffBlock

def compute_levenshtein_blocks(seq1, seq2, max_complexity=1e8):
    """Compute the Levenshtein blocks of insertion, deletion, replacement.
    """
    # TODO: better method for dealing with long sequences?
    l1, l2 = len(seq1), len(seq2)
    if l1 * l2 > max_complexity:
        return [('change', (0, l1), (0, l2))]

    def block_format(op, s1, e1, s2, e2):
        if op == "delete":
            return (op, (s1, e1 + 1), (s2, e2))
        if op == "insert":
            return (op, (s1, e1), (s2, e2 + 1))
        else:
            return (op, (s1, e1 + 1), (s2, e2 + 1))

    edits = Levenshtein.editops(seq1, seq2)
    if len(edits) == 0:
        return []
    bop, s1, s2 = edits[0]
    e1, e2 = s1, s2
    blocks = []
    for (op, _e1, _e2) in edits[1:]:
        continuity = any(
            [
                all([op == "delete", _e1 == e1 + 1, e2 == _e2]),
                all([op == "insert", _e1 == e1, _e2 == e2 + 1]),
                all([op == "replace", _e1 == e1 + 1, _e2 == e2 + 1]),
            ]
        )
        if op == bop and continuity:
            e1, e2 = _e1, _e2
        else:
            blocks.append(block_format(bop, s1, e1, s2, e2))
            bop, s1, s2 = op, _e1, _e2
            e1, e2 = s1, s2
    blocks.append(block_format(bop, s1, e1, s2, e2))
    return blocks


def merge_subblocks(subblocks):
    s1_location = Location(
        min([b.s1_location.start for b in subblocks]),
        max([b.s1_location.end for b in subblocks]),
        sequence=subblocks[0].s1_location.sequence,
    )
    s2_location = Location(
        min([b.s2_location.start for b in subblocks]),
        max([b.s2_location.end for b in subblocks]),
        sequence=subblocks[0].s2_location.sequence,
    )
    if len(s1_location) == len(s2_location):
        return DiffBlock("replace", s1_location, s2_location)
    else:
        return DiffBlock("change", s1_location, s2_location)


def merge_blocs_by_location(blocks, max_blocks, max_span, reference="s2"):
    while 1:
        for i in range(len(blocks) - max_blocks):
            subblocks = blocks[i : i + max_blocks]
            b1, b2 = subblocks[0], subblocks[-1]
            if reference == "s2":
                span = b2.s2_location.end - b1.s2_location.start
            else:
                span = b2.s1_location.end - b1.s1_location.start
            if span < max_span:
                new_block = merge_subblocks(subblocks)
                blocks = blocks[:i] + [new_block] + blocks[i + max_blocks :]
                break
        else:
            break
    return blocks


def merge_successive_blocks(blocks, change_gap=10, reference="s2"):
    while 1:
        for i in range(len(blocks) - 1):
            b1, b2 = blocks[i], blocks[i + 1]
            if "equal" in [b1.operation, b2.operation]:
                continue
            if "change" not in [b1.operation, b2.operation]:
                continue
            if reference == "s2":
                gap = b2.s2_location.start - b1.s2_location.end
            else:
                gap = b2.s1_location.start - b1.s1_location.end
            if gap < change_gap:
                new_block = merge_subblocks([b1, b2])
                blocks = blocks[:i] + [new_block] + blocks[i + 2 :]
                break
        else:
            break
    return blocks


def compute_sorted_blocks(blocks, reference="s2"):
    def sort_key(block):
        if reference == "s2":
            return block.s2_location.to_tuple()
        else:
            return block.s1_location.to_tuple()

    return sorted(blocks, key=sort_key)


def get_optimal_common_blocks(common_blocks):
    common_blocks = deepcopy(common_blocks)
    remarks = []

    # Make so that there is never an anitsense block in s1 and a + block in s2.
    # If it is so, flip the block in s2. It will become antisense and be later
    # removed
    for block_name, data in common_blocks.items():
        locations = data["locations"]
        s1_strands = [strand for (s, (_, _, strand)) in locations if s == "s1"]
        if 1 not in s1_strands:
            for i, location in enumerate(locations):
                seq, (start, end, strand) = location
                if seq == "s2":
                    locations[i] = (seq, (start, end, -strand))

    # Remove every antisense blocks now. For the ones in s2, log this
    # with a remark.
    for block_name, data in common_blocks.items():
        locations = data["locations"]
        for i, location in enumerate(locations):
            (seq, (start, end, strand)) = location
            if seq == "s2" and (strand == -1):
                locations.remove(location)
                _, (start1, end1, strand1) = locations[0]
                remarks.append(
                    DiffBlock(
                        "reverse",
                        s1_location=Location(start1, end1, strand1),
                        s2_location=Location(start, end),
                    )
                )

    # We start the structure that will be returned in the end
    blocks_in_seqs = {
        seq: sorted(
            [
                (start, end, bname)
                for bname, data in common_blocks.items()
                for (s, (start, end, strand)) in data["locations"]
                if s == seq
            ]
        )
        for seq in ("s1", "s2")
    }

    # Identify blocks appearing only in one of the two sequences
    blocks_in_s1 = set(b[-1] for b in blocks_in_seqs["s1"])
    blocks_in_s2 = set(b[-1] for b in blocks_in_seqs["s2"])
    uniblocks = (blocks_in_s1.union(blocks_in_s2)).difference(
        blocks_in_s1.intersection(blocks_in_s2)
    )

    # Remove blocks appearing only in one of the two sequences
    # as they are useless for sequences comparison
    # this should be very rare but you never know.
    for block_list in blocks_in_seqs.values():
        for b in block_list:
            if b[-1] in uniblocks:
                block_list.remove(b)

        for b1, b2 in zip(block_list, block_list[1:]):
            start1, end1, __name1 = b1
            start2, end2, __name2 = b2
            if end2 <= end1:
                block_list.remove(b2)

    # If a block appears several time in a sequence (self-homology)
    # give unique names to each occurence: block_1, block_1*, etc.
    blocks_in_seqs_dicts = dict(s1={}, s2={})
    for seq, blocks_list in list(blocks_in_seqs.items()):
        seen_blocks = set()
        for i, (start, end, block_name) in enumerate(blocks_list):
            while block_name in seen_blocks:
                block_name = block_name + "*"
            blocks_list[i] = start, end, block_name
            blocks_in_seqs_dicts[seq][block_name] = dict(
                rank=i, location=(start, end)
            )
            seen_blocks.add(block_name)

    # Find and retain the largest sequence of blocks which is in the right order
    # in both sequences. We will remove every other blocks.
    # We do that by looking for the longest path in a graph

    if len(blocks_in_seqs_dicts["s2"]) < 2:
        retained_blocks = list(blocks_in_seqs_dicts["s2"])
    else:
        s1_dict = blocks_in_seqs_dicts["s1"]
        graph = nx.DiGraph(
            [
                (b1, b2)
                for b1, data1 in blocks_in_seqs_dicts["s2"].items()
                for b2, data2 in blocks_in_seqs_dicts["s2"].items()
                if (b2 in s1_dict)
                and (b1 in s1_dict)
                and (s1_dict[b2]["rank"] > s1_dict[b1]["rank"])
                and (data2["rank"] > data1["rank"])
            ]
        )
        retained_blocks = nx.dag_longest_path(graph)

    # remove any "misplaced" block that is not in the retained list.
    # log a remark for the ones in s2.
    for seq in ("s1", "s2"):
        blocks_list = blocks_in_seqs[seq]
        for block in list(blocks_list):  # copy cause we will remove elements
            start, end, block_name = block
            if block_name not in retained_blocks:
                blocks_list.remove(block)
                if seq == "s2":
                    new_block_name = block_name.strip("*")
                    s1_blocks = blocks_in_seqs_dicts["s1"]
                    s1_loc = s1_blocks[new_block_name]["location"]
                    start1, end1 = s1_loc
                    if (
                        len(remarks)
                        and (start == remarks[-1].s2_location.end)
                        and (start1 == remarks[-1].s1_location.end)
                    ):
                        remarks[-1].s1_location.end = end1
                        remarks[-1].s2_location.end = end
                    else:
                        remarks.append(
                            DiffBlock(
                                "transpose",
                                s1_location=Location(start1, end1),
                                s2_location=Location(start, end),
                            )
                        )

    # Reduce blocks when there is overlap
    blocks_to_reduce = {}
    for seq in ("s1", "s2"):
        blocks_list = blocks_in_seqs[seq]
        for b1, b2 in zip(blocks_list, blocks_list[1:]):
            start1, end1, block_name1 = b1
            start2, end2, __block_name2 = b2
            diff = end1 - start2
            if diff > 0:
                if block_name1 not in blocks_to_reduce:
                    blocks_to_reduce[block_name1] = 0
                blocks_to_reduce[block_name1] = max(
                    blocks_to_reduce[block_name1], diff
                )

    for seq in ("s1", "s2"):
        blocks_list = blocks_in_seqs[seq]
        for i, (start, end, block_name) in enumerate(blocks_list):
            if block_name in blocks_to_reduce:
                diff = blocks_to_reduce[block_name]
                blocks_list[i] = (start, end - diff, block_name)

    return blocks_in_seqs, remarks



from .biotools import reverse_complement


def change(seq, start, end, change):
    """Return the sequence with ``seq[start:end]`` replaced by ``change``"""
    return seq[:start] + change + seq[end:]


def insert(seq, pos, inserted):
    """Return the sequence with ``inserted`` inserted, starting at index 'pos'
    """
    return seq[:pos] + inserted + seq[pos:]


def delete(seq, pos, deletions):
    """Return the sequence with a number of deletions from position pos."""
    return seq[:pos] + seq[pos + deletions :]


def reverse(seq, start, end):
    """Return the sequence with segment seq[start:end] reverse-complemented."""
    return seq[:start] + reverse_complement(seq[start:end]) + seq[end:]


def move(seq, start, end, diff):
    """Move a subsequence by "diff" nucleotides the left or the right."""
    sub = seq[start:end]
    if diff > 0:
        return seq[:start] + seq[end : end + diff] + sub + seq[end + diff :]
    else:
        return (
            seq[: start + diff] + sub + seq[start + diff : start] + seq[end:]
        )


def swap(seq, pos1, pos2):
    """Return a new sequence with segments at position pos1 and pos2 swapped.

    pos1, pos2 are both of the form (start1, end1), (start2, end2)
    """
    (start1, end1), (start2, end2) = sorted([pos1, pos2])
    return (
        seq[:start1]
        + seq[start2:end2]
        + seq[end1:start2]
        + seq[start1:end1]
        + seq[end2:]
    )


def copy(seq, start, end, new_start):
    """Return the sequence with segment [start, end] also copied elsewhere,
    starting in new_start."""
    return insert(seq, new_start, seq[start:end])



from .biotools import reverse_complement
from Bio.SeqFeature import SeqFeature, FeatureLocation


class Location:
    def __init__(
        self, start, end, strand=None, sequence=None, sequence_id=None
    ):

        self.start = start
        self.end = end
        self.strand = strand
        self.sequence = sequence
        self.sequence_id = sequence_id

    def extract_sequence(self, sequence=None):
        """Return the subsequence read at the given location.

        If sequence is None, ``self.sequence`` is used.
        """
        if sequence is None:
            sequence = self.sequence
        if hasattr(sequence, "seq"):
            sequence = str(sequence.seq)
        result = sequence[self.start : self.end]
        if self.strand == -1:
            return reverse_complement(result)
        else:
            return result

    def __repr__(self):
        """Represent"""
        result = "%d-%d" % (self.start, self.end)
        if self.strand is not None:
            result += {1: "(+)", -1: "(-)", 0: ""}[self.strand]
        if self.sequence_id is not None:
            result = self.sequence_id + "|" + result
        return result

    def __len__(self):
        """Size of the location"""
        return abs(self.end - self.start)

    def to_tuple(self):
        return self.start, self.end, self.strand

    def to_biopython_location(self):
        """Return a Biopython FeatureLocation equivalent to the location."""
        start, end, strand = [
            None if e is None else int(e)
            for e in [self.start, self.end, self.strand]
        ]
        return FeatureLocation(start, end, strand)

    def to_biopython_feature(self, feature_type="misc_feature", **qualifiers):
        """Return a Biopython SeqFeature with same location and custom
        qualifiers."""
        return SeqFeature(
            self.to_biopython_location(),
            type=feature_type,
            qualifiers=qualifiers,
        )



"""Defines central class BlockFinder."""
import itertools
from collections import OrderedDict
from copy import deepcopy


from dna_features_viewer import BiopythonTranslator
import matplotlib.pyplot as plt
import matplotlib.cm as cm

from ..biotools import annotate_record

from .commonblocks_tools import (
    format_sequences_as_dicts,
    select_common_blocks,
    find_homologies_between_sequences,
)

# TODO: Simplify the code by using the new Location class in Location.py


class CommonBlocks:
    """Class to represent a set of common blocks from different sequences.

    Create with ``CommonBlocks.from_sequences``:

    >>> common_blocks = CommonBlocks.from_sequences({'s1': 'ATGC...'})

    Parameters
    ----------

    common_blocks
      A dictionnary of the sequences to compare, of the form
      {sequence_name: ATGC_sequence_string} or a list of records, all with
      different IDs.

    records
      A dictionnary of the biopython records of the sequences
      {record_id: record}
    """

    def __init__(self, common_blocks, records):
        """Initialize, compute best blocks."""
        self.common_blocks = common_blocks
        self.records = records

    @staticmethod
    def from_sequences(
        sequences,
        block_selection_method="most_coverage_first",
        include_self_homologies=True,
        min_block_size=80,
        max_block_size=None,
    ):
        sequences_dict, records_dict = format_sequences_as_dicts(sequences)
        homologies_dict = find_homologies_between_sequences(
            sequences_dict,
            min_size=min_block_size,
            max_size=max_block_size,
            include_self_homologies=include_self_homologies,
        )
        common_blocks = select_common_blocks(
            homologies_dict, sequences_dict, min_size=min_block_size,
            method=block_selection_method
        )
        return CommonBlocks(common_blocks=common_blocks, records=records_dict)

    def compute_unique_blocks(self):
        """Return a dictionnary listing unique blocks by sequence.

        The unique blocks are the blocks between the selected common blocks.

        The result is of the form {seq: [(start, end), (start2, end2), ...]}
        """

        unique_blocks = OrderedDict()
        for seqname, rec in self.sequences_with_annotated_blocks().items():
            blocks_locations = (
                [(0, 0)]
                + sorted(
                    [
                        (f.location.start, f.location.end)
                        for f in rec.features
                        if f.qualifiers.get("is_block", False)
                    ]
                )
                + [(len(rec), len(rec))]
            )
            unique_blocks[seqname] = [
                (end1, start2)
                for (_, end1), (start2, _) in zip(
                    blocks_locations, blocks_locations[1:]
                )
                if (start2 - end1) > 1
            ]
        return unique_blocks

    def common_blocks_to_csv(self, target_file=None):
        """Write the common blocks into a CSV file.

        If a target CSV file is provided the result is written to that file.
        Otherwise the result is returned as a string.


        The columns of the CSV file are "block", "size", "locations", and
        sequence.

        """
        csv_content = "\n".join(
            ["block;size;locations;sequence"]
            + [
                ";".join(
                    [
                        block_name,
                        str(len(data["sequence"])),
                        " ".join(
                            [
                                "%s(%d, %d, %d)" % (cst, start, end, strand)
                                for (cst, (start, end, strand)) in data[
                                    "locations"
                                ]
                            ]
                        ),
                        data["sequence"],
                    ]
                )
                for block_name, data in self.common_blocks.items()
            ]
        )
        if target_file:
            with open(target_file, "w+") as f:
                f.write(csv_content)
        else:
            return csv_content

    def common_blocks_records(self):
        """Return all common blocks as a list of Biopython records.
        """
        if self.records is None:
            raise ValueError("")
        records = []
        for block_name, data in self.common_blocks.items():
            cst, (start, end, strand) = data["locations"][0]
            record = self.records[cst][start:end]
            if strand == -1:
                record = record.reverse_complement()
            record.id = record.name = block_name
            records.append(record)
        return records

    def unique_blocks_records(self, target_file=None):
        """Return all unique blocks as a list of Biopython records."""
        if self.records is None:
            raise ValueError("")
        records = []
        for seqname, locations in self.compute_unique_blocks().items():
            for i, (start, end) in enumerate(locations):
                record = self.records[seqname][start:end]
                record.id = "%s_%03d" % (seqname, i)
                records.append(record)
        return records

    def sequences_with_annotated_blocks(self, colors="auto"):
        """Return a list of Biopython records representing the sequences
        with annotations indicating the common blocks

        Parameter ``colors`` is either a list of colors or "auto" for the
        default.
        """
        records = deepcopy(self.records)
        if colors == "auto":
            colors = itertools.cycle(
                [cm.Paired(0.21 * i % 1.0) for i in range(30)]
            )
        blocks_and_colors = zip(self.common_blocks.items(), colors)
        for (name, data), color in blocks_and_colors:
            for (seqname, location) in data["locations"]:
                annotate_record(
                    records[seqname],
                    location,
                    feature_type="misc_feature",
                    is_block=True,
                    label=name,
                    color=color,
                )
        return records

    def plot_common_blocks(
        self, colors="auto", axes=None, figure_width=10, ax_height=2
    ):
        """Plot the common blocks found on vertically stacked axes.

        The axes on which the plots are drawn are returned at the end.

        Parameters
        ----------

        colors
          Either a list of colors to use for blocks or "auto" for the default.

        axes
          A list of matplotlib axes on which to plot, or None for new axes

        figure_width
          Width of the final figure in inches

        ax_eight
          Height of each plot
        """

        translator = BiopythonTranslator()
        records = self.sequences_with_annotated_blocks(colors=colors)
        if axes is None:
            fig, axes = plt.subplots(
                len(self.records),
                1,
                facecolor="white",
                sharex=True,
                figsize=(figure_width, ax_height * len(self.records)),
            )
        else:
            fig = axes[0].figure
        for (ax, (seqname, record)) in zip(axes, records.items()):
            gr_record = translator.translate_record(record)
            gr_record.plot(
                ax,
                x_lim=(0, max([len(rec) for rec in self.records.values()])),
                with_ruler=(ax == axes[-1]),
            )
            ax.set_title(seqname, loc="left", fontdict=dict(weight="bold"))
        fig.tight_layout()
        return axes

    def copy_features_between_common_blocks(self, inplace=False):
        def extract_subrecord(record, location):
            start, end, strand = location
            record = record[start:end]
            if strand == -1:
                record = record.reverse_complement()
            return record

        def extract_features(record, offset, reverse=False):
            if reverse:
                record = record.reverse_complement()
            new_features = [deepcopy(f) for f in record.features]
            for f in new_features:
                f.qualifiers["original_record"] = record.id
            for f in new_features:
                f.location += offset
            return new_features

        if inplace:
            records = self.records
        else:
            records = deepcopy(self.records)
        for data in self.common_blocks.values():
            locations = data["locations"]
            subrecords = {
                rec_id: extract_subrecord(records[rec_id], location)
                for rec_id, location in data["locations"]
            }
            for l1, l2 in itertools.combinations(locations, 2):
                for ((id1, loc1), (id2, __loc2)) in ((l1, l2), (l2, l1)):
                    start1, __end1, strand1 = loc1
                    # start2, end2, strand2 = loc2
                    records[id1].features += extract_features(
                        subrecords[id2], offset=start1, reverse=(strand1 == -1)
                    )
        return records



import os
from geneblocks import DiffBlocks, load_record

seq_1 = load_record(os.path.join("sequences", "sequence1.gb"))
seq_2 = load_record(os.path.join("sequences", "sequence2.gb"))

diff_blocks = DiffBlocks.from_sequences(seq_1, seq_2)
ax1, ax2 = diff_blocks.plot(figure_width=8)
ax1.figure.savefig("diff_blocks.png", bbox_inches='tight')



import tempfile
import subprocess

import numpy as np

try:
    from Bio.Seq import Seq
    from Bio.SeqRecord import SeqRecord
    from Bio.Alphabet import DNAAlphabet
    from Bio.SeqFeature import SeqFeature, FeatureLocation
    from Bio import SeqIO

    BIOPYTHON_AVAILABLE = True
except ImportError:
    BIOPYTHON_AVAILABLE = False

complements_dict = {"A": "T", "T": "A", "C": "G", "G": "C"}


def random_dna_sequence(length, probas=None, seed=None):
    """Return a random DNA sequence ("ATGGCGT...") with the specified length.

    Parameters
    ----------

    length
      Length of the DNA sequence.

    proba
      Frequencies for the different nucleotides, for instance
      ``probas={"A":0.2, "T":0.3, "G":0.3, "C":0.2}``.
      If not specified, all nucleotides are equiprobable (p=0.25).

    seed
      The seed to feed to the random number generator. When a seed is provided
      the random results depend deterministically on the seed, thus enabling
      reproducibility

    """
    if seed is not None:
        np.random.seed(seed)
    if probas is None:
        sequence = np.random.choice(list("ATCG"), length)
    else:
        bases, probas = zip(*probas.items())
        sequence = np.random.choice(bases, length, p=probas)
    return "".join(sequence)


def load_record(filename, linear=True, name="id", upperize=True):
    formt = "genbank" if filename.endswith(("gb", "gbk")) else "fasta"
    record = SeqIO.read(filename, formt)
    if upperize:
        record.seq = record.seq.upper()
    record.linear = linear
    if name != "id":
        record.id = name
    record.name = record.id.replace(" ", "_")[:20]

    return record


def complement(sequence):
    return "".join(complements_dict[c] for c in sequence)


def reverse_complement(sequence):
    return complement(sequence)[::-1]


def sequence_to_record(sequence, record_id=None, name="unnamed", features=()):
    if not BIOPYTHON_AVAILABLE:
        raise ImportError("Creating records requires Biopython installed.")
    return SeqRecord(
        Seq(sequence, alphabet=DNAAlphabet()),
        name=name,
        id=record_id,
        features=list(features),
    )


def annotate_record(
    seqrecord, location="full", feature_type="feature", margin=0, **qualifiers
):
    """Add a feature to a Biopython SeqRecord.

    Parameters
    ----------

    seqrecord
      The biopython seqrecord to be annotated.

    location
      Either (start, end) or (start, end, strand). (strand defaults to +1)

    feature_type
      The type associated with the feature

    margin
      Number of extra bases added on each side of the given location.

    qualifiers
      Dictionnary that will be the Biopython feature's `qualifiers` attribute.
    """

    if not BIOPYTHON_AVAILABLE:
        raise ImportError("Creating records requires Biopython installed.")

    if location == "full":
        location = (margin, len(seqrecord) - margin)

    strand = location[2] if len(location) == 3 else 1
    seqrecord.features.append(
        SeqFeature(
            FeatureLocation(location[0], location[1], strand),
            qualifiers=qualifiers,
            type=feature_type,
        )
    )


def sequences_differences_array(seq1, seq2):
    """Return an array [0, 0, 1, 0, ...] with 1s for sequence differences.

    seq1, seq2 should both be ATGC strings.
    """
    if len(seq1) != len(seq2):
        raise ValueError(
            "Only use on same-size sequences (%d, %d)" % (len(seq1), len(seq2))
        )
    arr1 = np.fromstring(seq1, dtype="uint8")
    arr2 = np.fromstring(seq2, dtype="uint8")
    return arr1 != arr2


def sequences_differences(seq1, seq2):
    """Return the number of nucleotides that differ in the two sequences.

    seq1, seq2 should be strings of DNA sequences e.g. "ATGCTGTGC"
    """
    return sequences_differences_array(seq1, seq2).sum()


from ..Location import Location
from ..biotools import sequences_differences

class DiffBlock:
    """Class to represent a segment that differs between sequences.


    Parameters
    ----------
    operation
      One of "insert", "delete", "replace", or "equal".
    
    s1_location
      The Location(start, end) of the region in sequence sequence s1.
    
    s2_location
      The Location(start, end) of the region in sequence sequence s2.
    
    """

    def __init__(self, operation, s1_location, s2_location):
        self.operation = operation
        self.s1_location = s1_location
        self.s2_location = s2_location

    def to_feature(self, sequence="s2"):
        s1_length, s2_length = len(self.s1_location), len(self.s2_location)
        max_length = max([s1_length, s2_length])

        if sequence == "s1":
            return DiffBlock(
                self.operation, self.s2_location, self.s1_location
            ).to_feature()

        if self.operation == "insert":
            if max_length <= 7:
                label = "+%s" % self.s2_location.extract_sequence()
            else:
                label = "+ %d nuc." % s2_length
        elif self.operation == "delete":
            if max_length <= 7:
                label = "-%s" % self.s1_location.extract_sequence()
            else:
                label = "- %d nuc." % s1_length
        elif self.operation == "replace":
            if max([s1_length, s2_length]) <= 6:
                label = "%s ➤ %s" % (
                    self.s1_location.extract_sequence(),
                    self.s2_location.extract_sequence(),
                )
            else:
                sub_s1 = self.s1_location.extract_sequence()
                sub_s2 = self.s2_location.extract_sequence()
                diffs = sequences_differences(sub_s1, sub_s2)
                label = "%d/%dn replaced" % (diffs, s1_length)

        elif self.operation == "change":
            if max([s1_length, s2_length]) <= 6:
                label = "%s ➤ %s" % (
                    self.s1_location.extract_sequence(),
                    self.s2_location.extract_sequence(),
                )
            else:
                label = "%sn ➤ %sn change" % (s1_length, s2_length)
        elif self.operation == "reverse":
            label = "was reversed at %d-%d" % (
                self.s1_location.start,
                self.s1_location.end,
            )
        elif self.operation == "transpose":
            label = "was at %d-%d" % (
                self.s1_location.start,
                self.s1_location.end,
            )
        elif self.operation == "equal":
            label = "Equal"

        return self.s2_location.to_biopython_feature(
            feature_type="diff_" + self.operation, label=label
        )

    def __str__(self):
        """Represent a diffbloc, for instance: "insert 10-30|105-105" """
        return "%s %s|%s" % (
            self.operation,
            self.s1_location,
            self.s2_location,
        )

    def __repr__(self):
        return str(self)



from copy import deepcopy

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from ..Location import Location
from ..biotools import sequence_to_record
from ..CommonBlocks import CommonBlocks

from .DiffBlock import DiffBlock
from .DiffRecordTranslator import DiffRecordTranslator
from .diffblocks_tools import (
    compute_levenshtein_blocks,
    get_optimal_common_blocks,
    merge_blocs_by_location,
    merge_successive_blocks,
    compute_sorted_blocks,
)


class DiffBlocks:
    """Class to generate and represent DiffBlocks.

    Usage:

    >>> DiffBlocks.from_sequences(s1, s2)
    """

    def __init__(self, s1, s2, blocks):
        self.s1 = s1
        self.s2 = s2
        self.blocks = blocks

    @staticmethod
    def from_sequences(s1, s2, blast_over=500, max_complexity=1e8):
        """Create DiffBlocks by comparing two sequences
        
        Parameters
        ----------

        s1, s2
          Two sequences, either "ATGC..." strings or Biopython records.

        blast_over
          A blast will be triggered to accelerate homology finding if
          len(s1) + len(s2) > blast_over.
        
        max_complexity
          If len(s1) * len(s2) is over max_complexity, no analysis is done and
          s1 is just labeled as a "change" of s2 (useful internally during the
          recursions of this method)
        
        """
        # Note: the sequences will always be upperized before they are
        # compared. however we also need to keep the initial sequences to
        # create the final blocks (possibly with upper/lowercase nucleotides)

        # If the sequences are records,  convert to string

        seq_s1 = str(s1.seq) if hasattr(s1, "seq") else str(s1)
        seq_s2 = str(s2.seq) if hasattr(s2, "seq") else str(s2)

        # Simple case to eliminate the trivial case of equality

        if seq_s1.upper() == seq_s2.upper():
            return DiffBlocks(s1, s2, [])

        # If the sequences are too big for straight-on Levenshtein, first
        # find the large sub-blocks that are identical, and the ones that
        # differ.

        if (blast_over is not None) and (len(s1) + len(s2)) > blast_over:

            diffblocks = []

            # Use CommonBlocks to find all big common blocks

            sequences = {"s1": s1, "s2": s2}
            common_blocks = CommonBlocks.from_sequences(
                sequences,
                min_block_size=100,
                include_self_homologies=False,
                block_selection_method="larger_first",
            ).common_blocks
            blocks_in_seqs, remarks = get_optimal_common_blocks(common_blocks)

            # First, each common block is added as an "equal" diffblock

            for b1, b2 in zip(blocks_in_seqs["s1"], blocks_in_seqs["s2"]):
                diffblocks.append(
                    DiffBlock(
                        "equal",
                        s1_location=Location(*b1[:2], sequence=s1),
                        s2_location=Location(*b2[:2], sequence=s2),
                    )
                )

            # for sequence in s1, s2, complete the sequence's list of blocks
            # with a (0, 0, "START") on the left, (L, L, "END") on the right.
            for seq, blocks in blocks_in_seqs.items():
                blocks_in_seqs[seq] = (
                    [(0, 0, "START")]
                    + blocks_in_seqs[seq]
                    + [(len(sequences[seq]), len(sequences[seq]), "END")]
                )
            for i in range(len(blocks_in_seqs["s2"]) - 1):
                _, end1, _ = blocks_in_seqs["s1"][i]
                next_start1, _, _ = blocks_in_seqs["s1"][i + 1]
                _, end2, _ = blocks_in_seqs["s2"][i]
                next_start2, _, _ = blocks_in_seqs["s2"][i + 1]
                if next_start2 < end2:
                    subdiffblocks = [
                        DiffBlock(
                            "delete",
                            s1_location=Location(
                                end1, next_start1, sequence=s1
                            ),
                            s2_location=Location(
                                next_start2, next_start2, sequence=s2
                            ),
                        )
                    ]
                else:
                    subsequence_1 = s1[end1:next_start1]
                    subsequence_2 = s2[end2:next_start2]
                    subdiffblocks = DiffBlocks.from_sequences(
                        subsequence_1,
                        subsequence_2,
                        blast_over=None,
                        max_complexity=max_complexity,
                    )
                    for block in subdiffblocks.blocks:
                        block.s1_location.start += end1
                        block.s1_location.end += end1
                        block.s1_location.sequence = s1
                        block.s2_location.start += end2
                        block.s2_location.end += end2
                        block.s2_location.sequence = s2
                diffblocks += subdiffblocks.blocks
            diffblocks = [
                b
                for b in diffblocks
                if len(b.s1_location) or len(b.s2_location)
            ]
            sorted_blocks = compute_sorted_blocks(diffblocks + remarks)
            return DiffBlocks(s1, s2, sorted_blocks)

        s1_std = str(s1.seq if hasattr(s1, "seq") else s1).upper()
        s2_std = str(s2.seq if hasattr(s2, "seq") else s2).upper()
        levenshtein_blocks = compute_levenshtein_blocks(
            s1_std, s2_std, max_complexity=max_complexity
        )
        blocks = [
            DiffBlock(
                operation,
                Location(s1s, s1e, sequence=s1),
                Location(s2s, s2e, sequence=s2),
            )
            for operation, (s1s, s1e), (s2s, s2e) in levenshtein_blocks
        ]
        return DiffBlocks(s1, s2, blocks)

    def merged(self, blocks_per_span=(3, 600), change_gap=100, reference="s2"):
        blocks = [
            b
            for b in self.blocks
            if b.operation not in ["reverse", "transpose"]
        ]
        remarks = [
            b for b in self.blocks if b.operation in ["reverse", "transpose"]
        ]
        if blocks_per_span is not None:
            max_blocks, span = blocks_per_span
            blocks = merge_blocs_by_location(
                blocks=blocks,
                max_blocks=max_blocks,
                max_span=span,
                reference=reference,
            )
        if change_gap is not None:
            blocks = merge_successive_blocks(
                blocks=blocks, change_gap=change_gap, reference="s2"
            )
        blocks = compute_sorted_blocks(blocks + remarks)
        return DiffBlocks(s1=self.s1, s2=self.s2, blocks=blocks)

    def sort_blocks(self):
        self.blocks = compute_sorted_blocks(self.blocks)

    def diffs_as_features(self, sequence="s2"):
        return [block.to_feature(sequence=sequence) for block in self.blocks]

    def plot(
        self,
        translator_class="default",
        separate_axes=True,
        sequence="s2",
        **plot_kw
    ):

        if translator_class == "default":
            translator_class = DiffRecordTranslator
        translator = translator_class()
        record = deepcopy(self.s2 if sequence == "s2" else self.s1)
        if not hasattr(record, "features"):
            record = sequence_to_record(record)

        diff_features = self.diffs_as_features(sequence=sequence)

        if separate_axes:
            gr_record = translator.translate_record(record)
            record.features = diff_features
            gr_diffrecord = DiffRecordTranslator().translate_record(record)
            width = plot_kw.get("figure_width", 8)
            if "axes" in plot_kw:
                ax1, ax2 = plot_kw["axes"]
                fig = ax1.figure
            else:
                fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(width, 6))
            plot_kw['annotate_inline'] = plot_kw.get('annotate_inline', True)
            _, stats1 = gr_record.plot(ax=ax1, **plot_kw)
            _, stats2 = gr_diffrecord.plot(ax=ax2, with_ruler=False, **plot_kw)
            max_level_1 = max(
                [0] + [v["annotation_y"] for v in stats1[1].values()]
            )
            max_features_1 = max(
                [0] + [v["feature_y"] for v in stats1[1].values()]
            )
            max_level_2 = max(
                [0] + [v["annotation_y"] for v in stats2[1].values()]
            )
            n_levels = max_level_1 + max_level_2
            if max_level_1 and max_level_2:
                plt.close(fig)

                ##

                gs = gridspec.GridSpec(n_levels, 1)
                fig = plt.figure(
                    figsize=(width, 1 + 0.6 * n_levels), facecolor="w"
                )
                ax1 = fig.add_subplot(gs[:max_level_1])
                ax2 = fig.add_subplot(gs[max_level_1:])
                _, stats1 = gr_record.plot(
                    ax=ax1, **plot_kw
                )
                _, stats2 = gr_diffrecord.plot(
                    ax=ax2, with_ruler=False, **plot_kw
                )
            fig.tight_layout()

            # fig.set_size_inches((width, 3 + 0.6 * n_levels))
            ax2.set_ylim(bottom=-0.5)
            ax2.invert_yaxis()
            for f in gr_diffrecord.features:
                ax1.fill_between(
                    [f.start, f.end],
                    y1=max_features_1 + 1,
                    y2=-1,
                    facecolor=f.color,
                    alpha=0.07,
                    zorder=1000,
                )
            return (ax1, ax2)

        else:
            record.features += diff_features
            gr_record = translator.translate_record(record)
            ax, _ = gr_record.plot(**plot_kw)
            return ax

    @staticmethod
    def reconstruct_sequences_from_blocks(blocks):
        s1, s2 = "", ""
        blocks = sorted(blocks, key=lambda b: b.s2_location.to_tuple())
        for block in blocks:
            if block.operation in ("equal", "replace", "change", "delete"):
                s1 = s1 + block.s1_location.extract_sequence()
            if block.operation in ("equal", "replace", "change", "insert"):
                s2 = s2 + block.s2_location.extract_sequence()
        return s1, s2

    def __str__(self):
        return ", ".join([str(b) for b in self.blocks])



""" geneblocks/__init__.py """

# __all__ = []

from .CommonBlocks import CommonBlocks
from .DiffBlocks import DiffBlocks, DiffRecordTranslator
from .biotools import load_record, random_dna_sequence, reverse_complement

from .version import __version__

__all__ = [
    "CommonBlocks",
    "DiffBlocks",
    "DiffRecordTranslator",
    "load_record",
    "random_dna_sequence",
    "reverse_complement",
    "__version__",
]



from geneblocks import DiffBlocks, CommonBlocks, random_dna_sequence
import geneblocks.sequence_modification_utils as smu
import matplotlib.pyplot as plt

# GENERATE 2 "SISTER" SEQUENCES FOR THE EXAMPLE
seq1 = random_dna_sequence(50000)
seq1 = smu.copy(seq1, 25000, 30000, 50000)

seq2 = seq1
seq2 = smu.insert(seq2, 39000, random_dna_sequence(100))
seq2 = smu.insert(seq2, 38000, random_dna_sequence(100))
seq2 = smu.reverse(seq2, 30000, 35000)
seq2 = smu.swap(seq2, (30000, 35000), (45000, 480000))
seq2 = smu.delete(seq2, 20000, 2000)
seq2 = smu.insert(seq2, 10000, random_dna_sequence(2000))
seq2 = smu.insert(seq2, 0, 1000*"A")

# FIND COMMON BLOCKS AND DIFFS
common_blocks = CommonBlocks.from_sequences({'seq1': seq1, 'seq2': seq2})
diff_blocks = DiffBlocks.from_sequences(seq1, seq2).merged()

# PLOT EVERYTHING
fig, axes = plt.subplots(3, 1, figsize=(15, 8))
common_blocks.plot_common_blocks(axes=axes[:-1])
diff_blocks.plot(ax=axes[-1], separate_axes=False)
axes[-1].set_xlabel("Changes in seq2 vs. seq1")
fig.savefig("complex_sequences.png", bbox_inches='tight')


from geneblocks import CommonBlocks
from geneblocks.biotools import reverse_complement, random_dna_sequence

C1, A, B, C2, C3, D, E, F = [random_dna_sequence(100 * L) for L in range(1, 9)]

sequences = {
    "a": C1 + A + C2,
    "b": B + D + C2 + C3,
    "c": E + C1 + C2,
    "d": C2 + C1 + F + C3,
    "e": C3 + reverse_complement(C2 + C1)
}

common_blocks = CommonBlocks.from_sequences(sequences)
axes = common_blocks.plot_common_blocks()
axes[0].figure.savefig("common_blocks.png", bbox_inches="tight")



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open('geneblocks/version.py').read()) # loads __version__

setup(
    name='geneblocks',
    version=__version__,
    author='Zulko',
    description='Library to compare DNA sequences (diff, common blocks, etc.)',
    long_description=open('pypi-readme.rst').read(),
    license='MIT',
    url='https://github.com/Edinburgh-Genome-Foundry/geneblocks',
    keywords="DNA sequence blocks diff synthetic-biology bioinformatics",
    packages=find_packages(exclude='docs'),
    install_requires=['numpy', 'Biopython', 'dna_features_viewer',
                      'networkx', 'python-Levenshtein'])



"""Algorithmic methods for the selection of common blocks in DiffBlocks

- select_common_blocks
  -
  - segments_difference
"""

import re
import tempfile
import subprocess
from collections import defaultdict, OrderedDict

import numpy as np
from ..biotools import reverse_complement, sequence_to_record


def format_sequences_as_dicts(sequences):
    """Standardize different formats into a single one.
    
    
    The ``sequences`` can bei either:
    - A list [('sequence_id', 'ATGC...'), ('sequence_2', ...)]
    - A list of Biopython records (all with different IDs)
    - A dict {'sequence_id': "ATGC..."}
    - A dict {'sequence_id': biopython_record}

    The output is a tuple (sequences_dict, records_dict), where
    - sequences_dict is of the form {'sequence_id': 'ATGC...'}
    - sequences_dict is of the form {'sequence_id': 'ATGC...'}
    """
    if isinstance(sequences, (list, tuple)):
        if hasattr(sequences[0], "seq"):
            # SEQUENCES = LIST OF RECORDS
            records_dict = OrderedDict(
                [(record.id, record) for record in sequences]
            )
            sequences_dict = OrderedDict(
                [(record.id, str(record.seq).upper()) for record in sequences]
            )
        else:
            # SEQUENCES = LIST OF ATGC STRINGS
            sequences_dict = OrderedDict(sequences)
            if isinstance(list(sequences_dict.values())[0], str):
                records_dict = OrderedDict(
                    [
                        (name, sequence_to_record(seq, name=name))
                        for name, seq in sequences_dict.items()
                    ]
                )
            else:
                records_dict = sequences
    elif hasattr(list(sequences.values())[0], "seq"):
        # SEQUENCES = DICT {SEQ_ID: RECORD}
        records_dict = OrderedDict(sorted(sequences.items()))
        sequences_dict = OrderedDict(
            [
                (record_id, str(record.seq).upper())
                for record_id, record in sequences.items()
            ]
        )
    else:
        # SEQUENCES = DICT {SEQ_ID: ATGC}
        sequences_dict = OrderedDict(sorted(sequences.items()))
        records_dict = OrderedDict(
            [
                (name, sequence_to_record(seq, name=name))
                for name, seq in sequences.items()
            ]
        )
    return sequences_dict, records_dict


def segments_difference(segment, subtracted):
    """Return the difference between segment (start, end) and subtracted.

    The result is a list containing either zero, one, or two segments of the
    form (start, end)

    Examples
    --------

    >>> segment=(10, 100), subtracted=(0, 85) => [(85, 100)]
    >>> segment=(10, 100), subtracted=(40, 125) => [(10, 40)]
    >>> segment=(10, 100), subtracted=(30, 55) => [(10, 30), (55, 100)]
    >>> segment=(10, 100), subtracted=(0, 150) => []
    
    
    """
    seg_start, seg_end = segment
    sub_start, sub_end = subtracted
    result = []
    if sub_start > seg_start:
        result.append((seg_start, min(sub_start, seg_end)))
    if sub_end < seg_end:
        result.append((max(seg_start, sub_end), seg_end))
    return sorted(list(set(result)))


def find_homologies_between_sequences(
    sequences, min_size=0, max_size=None, include_self_homologies=True
):
    """Return a dict listing the locations of all homologies between sequences.

    The result is a dict of the form below, where the sequence identifiers
    are used as keys
    
    >>> {
    >>>     'seq_1': {
    >>>        (start1, end1): [('seq2_5', _start, _end), ('seq_3', )...]
    >>>        (start2, end2): ...
    >>>     }
    >>> }

    Parameters
    ----------

    sequences
      A dict {'sequence_id': 'ATTGTGCAG...'}
    
    min_size, max_size
      Minimum and maximum size outside which homologies will be ignored.
    
    include_self_homologies
      If False, self-homologies will be removed from the list
    """

    # BLAST THE SEQUENCES USING NCBI-BLAST

    temp_fasta_path = tempfile.mktemp(".fa")

    with open(temp_fasta_path, "w+") as f:
        f.write(
            "\n\n".join(
                [
                    "> %s\n%s" % (name, seq.upper())
                    for name, seq in sequences.items()
                ]
            )
        )
    proc = subprocess.Popen(
        [
            "blastn",
            "-query",
            temp_fasta_path,
            "-subject",
            temp_fasta_path,
            "-perc_identity",
            "100",
            "-dust",
            "no",
            "-evalue",
            "100000000000",
            "-culling_limit",
            "10",
            "-ungapped",
            "-outfmt",
            "6 qseqid qstart qend sseqid sstart send",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    result, __blast_err = proc.communicate()

    # PARSE THE RESULT FROM BLAST

    parsing = [
        line.split("\t") for line in result.decode("utf-8").splitlines()
    ]
    homologies = {
        name: defaultdict(lambda *a: []) for name, seq in sequences.items()
    }

    # FILTER THE RESULTS (MIN_SIZE, MAX_SIZE, SELF-HOMOLOGIES)

    for query, qstart, qend, subject, sstart, send in parsing:
        is_self_homology = (query == subject) and (qstart != sstart)
        if is_self_homology and (not include_self_homologies):
            continue
        qstart, qend = int(qstart) - 1, int(qend)
        sstart, send = int(sstart) - 1, int(send)
        if qend - qstart < min_size:
            continue
        if (max_size is not None) and (qend - qstart > max_size):
            continue
        location = (subject, sstart, send)
        homologies[query][(qstart, qend)].append(location)
    return homologies


def count_homologies(matches, min_size):
    """Return a dict {(start, end): number_of_homologies_count}
    """
    homologies_counts = {}
    if len(matches) == 1:
        segment = list(matches.keys())[0]
        homologies_counts[segment] = 1
    matches_list = sorted(matches.keys())
    for i, match1 in enumerate(matches_list):
        for match2 in matches_list[i + 1 :]:
            segment = start, end = (match2[0], min(match1[1], match2[1]))
            if end < start:
                # The segment is empty, match1 and match2 as disjunct.
                break
            elif (end - start > min_size) and (
                segment not in homologies_counts
            ):
                homologies_counts[segment] = len(
                    [
                        matching
                        for (match_start, match_end) in matches_list
                        for matching in matches[(match_start, match_end)]
                        if match_start <= start <= end <= match_end
                    ]
                )
    return homologies_counts


def segment_with_most_homologies(
    homologies_counts, method="most_coverage_first"
):
    """Select the "best" segment, that should be selected next as a common
    block."""

    def segment_score(segment):
        if method == "most_coverage_first":
            factor = homologies_counts[segment]
        else:
            factor = 1
        start, end = segment
        return factor * (end - start)

    return max(
        [(0, (None, None))]
        + [(segment_score(segment), segment) for segment in homologies_counts]
    )


def select_common_blocks(
    homologies, sequences, min_size=0, method="most_coverage_first"
):
    """Select a collection of the largest common blocks, iteratively."""
    common_blocks = []
    homologies_counts = {
        seqname: count_homologies(
            matches=homologies[seqname], min_size=min_size
        )
        for seqname in sequences
    }

    # ITERATIVELY SELECT A COMMON BLOCK AND REMOVE THAT BLOCK FROM THE
    # homologies IN VARIOUS SEQUENCES, UNTIL THERE IS NO

    while True:

        # FIND THE HOMOLOGY WITH THE BEST OVERALL SCORE ACROSS ALL SEQS

        (best_score, (start, end)), seqname = max(
            [
                (
                    segment_with_most_homologies(
                        homologies_counts[seqname], method=method
                    ),
                    seqname,
                )
                for seqname in sequences
            ]
        )

        # IF NO HOMOLOGY WAS FOUND AT ALL, STOP

        if best_score == 0:
            break

        # FIND WHERE THE SELECTED SUBSEQUENCE APPEARS IN OTHER SEQUENCES.
        # AT EACH LOCATION, "EXTRUDE" THE SUBSEQUENCE FROM THE CURRENT
        # LOCATIONS IN homologies_counts

        best_subsequence = sequences[seqname][start:end]
        locations = []
        for seqname, sequence in sequences.items():
            seq_n_intersections = homologies_counts[seqname]

            # we look for both the subsequence and its reverse complement:
            for strand in [1, -1]:
                if strand == 1:
                    matches = re.finditer(best_subsequence, sequence)
                else:
                    matches = re.finditer(
                        reverse_complement(best_subsequence), sequence
                    )
                for match in matches:

                    # add the location to the list for this subsequence...
                    start, end = match.start(), match.end()
                    locations.append((seqname, (start, end, strand)))

                    # ...then subtract the location from the sequence's
                    # homologies list
                    match_as_segment = tuple(sorted([start, end]))
                    for intersection in list(seq_n_intersections.keys()):
                        score = seq_n_intersections.pop(intersection)
                        for diff in segments_difference(
                            intersection, match_as_segment
                        ):
                            diff_start, diff_end = diff
                            if diff_end - diff_start > min_size:
                                seq_n_intersections[diff] = score
        common_blocks.append((best_subsequence, locations))

    # REMOVE SELF-HOMOLOGOUS SEQUENCES

    common_blocks = [
        (seq, locations)
        for (seq, locations) in common_blocks
        if len(locations) >= 2
    ]

    # CREATE THE FINAL COMMON_BLOCKS_DICT

    common_blocks_dict = OrderedDict()
    if len(common_blocks) > 0:
        number_size = int(np.log10(len(common_blocks))) + 1
        for i, (sequence, locations) in enumerate(common_blocks):
            block_name = "block_%s" % (str(i + 1).zfill(number_size))
            common_blocks_dict[block_name] = {
                "sequence": sequence,
                "locations": locations,
            }

    return common_blocks_dict



import os
from geneblocks import CommonBlocks, load_record
from dna_features_viewer import BiopythonTranslator
import matplotlib.pyplot as plt

# LOAD THE TWO RECORDS

part_path = os.path.join("sequences", "features_transfer", "part.gb")
part = load_record(part_path, name="part")
plasmid_path = os.path.join(
    "sequences", "features_transfer", "plasmid_to_annotate.gb"
)
plasmid = load_record(plasmid_path, name="plasmid")


# TRANSFER THE FEATURES

blocks = CommonBlocks.from_sequences([part, plasmid])
new_records = blocks.copy_features_between_common_blocks(inplace=False)
annotated_plasmid = new_records["plasmid"]  # record with all features


# PLOT ALL RECORDS


fig, (ax0, ax1, ax2) = plt.subplots(3, 1, sharex=True, figsize=(7, 9))
ax0.set_title("Record 1: Annotated part", loc="left")
ax1.set_title("Record 2: Partially annotated plasmid", loc="left")
ax2.set_title("Record 2 (after) with all annotations", loc="left")


class SpecialBiopythonTranslator(BiopythonTranslator):
    def compute_feature_color(self, f):
        original_record = "".join(f.qualifiers.get("original_record", ""))
        return "#ffafaf" if (original_record == "part") else "#afafff"


tr = SpecialBiopythonTranslator()
tr.translate_record(part).plot(ax0, with_ruler=False, x_lim=(0, len(plasmid)))
tr.translate_record(plasmid).plot(ax1, with_ruler=False)
tr.translate_record(annotated_plasmid).plot(ax2, with_ruler=False)

fig.tight_layout()
fig.subplots_adjust(hspace=0.6)
fig.savefig("features_transfer.png", dpi=125)
